/* eslint-disable no-underscore-dangle */
import jwt from 'jsonwebtoken';
import { Request } from 'express';
import { Client } from '.';
import { AccountInterface } from '../models';

export default class Security {
  public client: Client;

  protected readonly keyBase: { key: string, iv: string, internal: string };

  constructor(client: Client) {
    this.client = client;
    this.keyBase = require(`${process.cwd()}/keys.json`);
  }

  get keys() {
    return {
      key: Buffer.from(this.keyBase.key, 'base64'),
      iv: Buffer.from(this.keyBase.iv, 'base64'),
      internal: Buffer.from(this.keyBase.internal, 'hex'),
    };
  }

  /**
   * Creates a new Bearer token.
   * @param _id The Mongoose Document property labeled ._id
   */
  public async createBearer(_id: string): Promise<string> {
    const account = await this.client.db.Account.findOne({ _id });
    if (!account) throw new Error(`Account [${_id}] cannot be found.`);
    return jwt.sign({ id: account.id }, this.keys.key, { issuer: 'Library of Code Inc | CSD' });
  }

  /**
   * If the bearer token is valid, will return the Account, else will return null.
   * @param bearer The bearer token provided.
   */
  public async checkBearer(bearer: string): Promise<null | AccountInterface> {
    try {
      const res: any = jwt.verify(bearer, this.keys.key, { issuer: 'Library of Code Inc | CSD' });
      const account = await this.client.db.Account.findOne({ _id: res.id });
      if (!account) return null;
      if (account.locked) return null;
      if (account.revokedBearers?.includes(bearer)) return null;
      return account;
    } catch {
      return null;
    }
  }

  /**
   * Returns the Bearer token, searches in headers and query.
   * @param req The Request object from Express.
   */
  public extractBearer(req: Request): string {
    if (req.headers.authorization && req.headers.authorization.split(' ')[0] === 'Bearer') {
      return req.headers.authorization.split(' ')[1];
    }
    if (req.query && req.query.token) {
      return req.query.token as string;
    }
    return '0000000000';
  }
}
